// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

package frc.robot;

import java.util.Optional;

import edu.wpi.first.wpilibj.DriverStation;
import edu.wpi.first.wpilibj.DriverStation.Alliance;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.Commands;
import edu.wpi.first.wpilibj2.command.button.CommandXboxController;
import frc.robot.constants.HoodConstants;
import frc.robot.constants.NeoMotorConstants;
import frc.robot.constants.OIConstants;
import frc.robot.subsystems.Climber;
import frc.robot.subsystems.Drivetrain;
import frc.robot.subsystems.Hood;
import frc.robot.subsystems.Indexer;
import frc.robot.subsystems.IntakePivot;
import frc.robot.subsystems.IntakeRollers;
import frc.robot.subsystems.Sensors;
import frc.robot.subsystems.Shooter;

public class RobotContainer {
    private Climber climber;
    private Drivetrain drivetrain;
    private Hood hood;
    private Indexer indexer;
    private IntakePivot intakePivot;
    private IntakeRollers intakeRollers;
    private Sensors sensors;
    private Shooter shooter;

    private CommandXboxController driver;
    private CommandXboxController operator;

    public RobotContainer() {
        climber = new Climber();
        drivetrain = new Drivetrain();
        hood = new Hood();
        indexer = new Indexer();
        intakePivot = new IntakePivot();
        intakeRollers = new IntakeRollers();
        sensors = new Sensors();
        shooter = new Shooter();

        driver = new CommandXboxController(OIConstants.kDriverUSB);
        operator = new CommandXboxController(OIConstants.kOperatorUSB);

        configureBindings();
    }

    private void configureBindings() {
        climber.setDefaultCommand(
            climber.stopClimber()
        );

        drivetrain.setDefaultCommand(drivetrain.drive(
            driver::getLeftX, 
            driver::getLeftY, 
            driver::getRightX, 
            () -> {
                Optional<Alliance> alliance = DriverStation.getAlliance();

                return alliance.isEmpty() || alliance.get() == Alliance.Blue ? false : true;
            }, 
            0.05, 
            true, 
            true
        ));

        hood.setDefaultCommand(
            hood.hoodDown()
        );

        indexer.setDefaultCommand(
            indexer.stopIndexer()
        );

        intakePivot.setDefaultCommand(
            intakePivot.intakeUp()
        );

        intakeRollers.setDefaultCommand(
            intakeRollers.stopIntake()
        );

        shooter.setDefaultCommand(
            shooter.manualShooter(
                driver::getRightTriggerAxis
            )
        );

        //driver inputs//
        driver.leftBumper().onTrue(shootGamePiece());

        driver.leftTrigger().whileTrue(autoIntake());

        driver.rightBumper().onTrue(ampShoot());
        
        //operator inputs//
        operator.povUp().onTrue(hood.hoodUp());

        operator.povDown().onTrue(autoClimb());

        operator.leftBumper().onTrue(ampLoad());
    }

    public Command getAutonomousCommand() {
        return Commands.print("No autonomous command configured");
    }

    //smart methods//
    private Command shootGamePiece() {
        return shooter.setSpeedShooter(NeoMotorConstants.kFreeSpeedRpm)
            .until(() -> shooter.atSetpoint())
            .withTimeout(2)
                .andThen(
                    shooter.setSpeedShooter(NeoMotorConstants.kFreeSpeedRpm)
                        .alongWith(indexer.runIndexer(1))
                        .until(() -> !sensors.getHoodBeamBreak() && !sensors.getIndexerBeamBreak())
                        .withTimeout(2)
                );
    }

    private Command ampLoad() {
        return indexer.runIndexer(1)
            .alongWith(shooter.setSpeedShooter(NeoMotorConstants.kFreeSpeedRpm / 4))
            .until(() -> sensors.getHoodBeamBreak())
            .withTimeout(2)
                .andThen(
                    hood.hoodUp()
                );
    }

    private Command ampShoot() {
        return shooter.setSpeedShooter(NeoMotorConstants.kFreeSpeedRpm)
        .until(() -> sensors.getHoodBeamBreak())
            .andThen(
                hood.hoodDown()
            );
    }

    private Command autoIntake() {
        return intakePivot.intakeDown()
            .until(() -> intakePivot.atSetpoint())
            .withTimeout(2)
            .andThen(
                intakePivot.intakeDown()
                    .alongWith(intakeRollers.runIntake(1))
                    .until(() -> sensors.getIndexerBeamBreak())
            );
    }

    private Command autoClimb() {
        return climber.runClimber(1)
            .until(() -> hood.getEncoder() <= HoodConstants.kHoodDownPosition)
            .withTimeout(2);
    }
}
// https://git.coldlightalchemist.com/NoahLacks63/2024again.git